// Copyright  2009 MergeIt, Aps.
//
// License LGPLv3+: GNU lesser LGPL version 3 or later <http://gnu.org/licenses/lgpl.html>.
// This is free software: you are free to change and redistribute it.
// There is NO WARRANTY, to the extent permitted by law.
//
//	This file is part of wiiscan.
//
//  wiiscan is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  any later version.
//
//  wiiscan is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with wiiscan.  If not, see <http://www.gnu.org/licenses/>.

#pragma warning(disable:4786) 

#define WIN32_LEAN_AND_MEAN		// Exclude rarely-used stuff from Windows headers
#include <windows.h>

#define  TOOLSFUN_QUIET_WIN32_SYSTEM // use WinExec instead of system  
#include "resource.h"
#include "systemtraysdk.h"
#include "wiiscan.h"
#include "process.h"

#define MAX_LOADSTRING 100
#define	WM_ICON_NOTIFY WM_APP+10

// Global Variables:
HINSTANCE hInst;					// current instance
char szTitle[MAX_LOADSTRING];		// The title bar text
char szWindowClass[MAX_LOADSTRING];	// The title bar text

CSystemTray  g_trayicon;
volatile int g_run=0;
const Configfile g_config(Wiiscan::DEFAULT_configfile);

string GetCurrentDir()
{
	FUNSTACK;
	char buff[16*1024];
	if(GetCurrentDirectory(16*1024,buff)==0) throw_("GetCurrentDirectory() failed");
	return tostring(buff);
}

string GetConfigStringMessage(const string& e)
{
	FUNSTACK;
	const pair<bool,string> p=g_config.GetConfig<string>(e);
	if (!p.first) Message("Warning...","Configuration file does not contain a <" + e + "=...> entry, or entry is empty");
	return p.second;
}

LRESULT CALLBACK About(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	FUNSTACK;
	switch (message) {
		case WM_INITDIALOG:	return TRUE;
		case WM_COMMAND:
			if (LOWORD(wParam) == IDOK || LOWORD(wParam) == IDCANCEL) {
				EndDialog(hDlg, LOWORD(wParam));
				return TRUE;
			}
			break;
	}
	return FALSE;
}

void RunWhiteboard(const bool killonly)
{		
	FUNSTACK;

	const string wb=strip(GetConfigStringMessage("whiteboard_software"),'"');
	if (wb!=""){
		const string e=tail(tail(tail(wb,':'),'/'),'\\');
		if (killonly){
			if (e!=""){
				DWORD id=GetProcessID(e);
				while(id!=0){
					KillProcessID(id);
					id=GetProcessID(e);
				}
			}
		} else {
			// fireup whiteboard software
			if (GetProcessCount(wb)>0) return;  
			else if (0!=SystemWin(wb)) throw_("could not launch executable <" + wb + ">");
		}
	}
}

void UpdateIconCallback(const int n)
{
	FUNSTACK;
	switch(n){
		case 0 : g_trayicon.SetIcon(IDI_SCAN_00); break;
		case 1 : g_trayicon.SetIcon(IDI_SCAN_01); break;
		case 2 : g_trayicon.SetIcon(IDI_SCAN_01); break;
		case 3 : g_trayicon.SetIcon(IDI_SCAN_02); break;
		case 4 : g_trayicon.SetIcon(IDI_SCAN_03); break;
		case 5 : g_trayicon.SetIcon(IDI_SCAN_03); break;
		case 6 : g_trayicon.SetIcon(IDI_SCAN_04); break;
		case 7 : g_trayicon.SetIcon(IDI_SCAN_05); break;
		case 8 : g_trayicon.SetIcon(IDI_SCAN_06); break;
		case 9 : g_trayicon.SetIcon(IDI_SCAN_07); break;
		default: g_trayicon.SetIcon(IDI_SCAN_07); break;
	}
}

DWORD WINAPI AutoIcon(LPVOID)
{
	int dir=0,state=-1;
	const string d=g_config.hasEntry("option_device") ? g_config.Get<string>("option_device") : Wiiscan::NINTENDO_DEV;
	const vector<string> known_adr=g_config("allowed_wiimote_adr");

	DeviceAutoClose<HANDLE,BOOL> hRadio(Wiiscan::RadioInfo("all",false),&CloseHandle);

	while(true){
		if (g_run==0) {
			dir=0;
			
			const int c=Wiiscan::isConnected(d,hRadio(),known_adr);

			if (c<0)                  {state=-1; g_trayicon.SetIcon(IDI_WIISCANICON_INV);}
			else if(c>0  && state!=1) {state=1; for(int i=0;i<=9;++i) {UpdateIconCallback(i); Sleep(50);}}
			else if(c==0 && state!=0) {state=0; for(int i=9;i>=0;--i) {UpdateIconCallback(i); Sleep(50);}}

			Sleep(100);			
		} else {	
			dir++;
			UpdateIconCallback(dir>=0 ? dir : -dir);
			if  (dir>=9) dir=-dir;
			Sleep(100);
		}
	}
}

void Register()
{
	g_run=1;

	const Configfile c(Wiiscan::DEFAULT_configfile);
	const int timeout=2000;
	const bool dbg=true;

	const int b=MessageBox(0,"Press the red-sync (hard-sync) button \non the wiimote\n\nand then press OK below...\n\nNOTE: pairing takes a minute or two, so be patient!","Wiimote pairing...",MB_OKCANCEL);
	if (b==2) return;

	Sleep(1000); // wait for wiimote to be ready

	BLUETOOTH_DEVICE_INFO bdi;
	memset(&bdi,0,sizeof(bdi));

	BLUETOOTH_RADIO_INFO radio;
	memset(&radio,0,sizeof(radio));
	
	const bool r=Wiiscan::Register("Nintendo RVL-CNT-01",c,"",timeout,dbg,&bdi,&radio);
	
	string s=r ?  "Device registeret successfully" : "Could not register device";
	s += "\n\nWiimote address: " + Wiiscan::ConvertAddress(bdi.Address);
	s += "\n\nBluetooth adapter address: " + Wiiscan::ConvertAddress(radio.address);
	if (r) s += "\nRemember to put the wiimote address into the <wiiscan.ini> file!";

	if (r) Sleep(60000);
	g_run=0;

	MessageBox(0,s.c_str(),"Wiimote pairing...",MB_OK);
}

void Unregister()
{
	g_run=2;
	const int b=MessageBox(0,"Sure you want to removed device?","Wiimote pairing...",MB_YESNO);
	if (b==7) {g_run=0; return;}

	const bool r=Wiiscan::RemoveDevice(Wiiscan::NINTENDO_DEV,false);
	if (r) g_trayicon.SetIcon(IDI_WIISCANICON_INV);
	g_run=0;

	if (r) MessageBox(0,"Device removed successfully!","Wiimote pairing...",MB_OK);
	else   MessageBox(0,"Could not remove device!","Wiimote pairing...",MB_OK);
}


LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	FUNSTACK;
	int wmId, wmEvent;
	switch (message) 
	{
		case WM_COMMAND:
			wmId    = LOWORD(wParam); 
			wmEvent = HIWORD(wParam); 
			// Parse the menu selections:
			switch (wmId) {
				case IDM_WHITEBOARD:      RunWhiteboard(false); break;
				case IDM_WHITEBOARD_KILL: RunWhiteboard(true); break; 
				case IDM_REGISTER:        Register(); break;
				case IDM_UNREGISTER:      Unregister(); break;
				case IDM_ABOUT:           DialogBox(hInst,(LPCTSTR)IDD_ABOUTBOX,hWnd,(DLGPROC)About); break;
				case IDM_EXIT:            DestroyWindow(hWnd); break;
				default:                  return DefWindowProc(hWnd, message, wParam, lParam);
			}
			break;
        case WM_ICON_NOTIFY: return g_trayicon.OnTrayNotification(wParam, lParam); 
		case WM_PAINT: 		 break;
		case WM_DESTROY:     PostQuitMessage(0); break;
		default:	         return DefWindowProc(hWnd, message, wParam, lParam);
   }
   return 0;
}

ATOM MyRegisterClass(HINSTANCE hInstance)
{
	FUNSTACK;
	WNDCLASSEX wcex;
	wcex.cbSize = sizeof(WNDCLASSEX); 

	wcex.style			= CS_HREDRAW | CS_VREDRAW;
	wcex.lpfnWndProc	= (WNDPROC)WndProc;
	wcex.cbClsExtra		= 0;
	wcex.cbWndExtra		= 0;
	wcex.hInstance		= hInstance;
	wcex.hIcon			= LoadIcon(hInstance, (LPCTSTR)IDI_WIISCANICON);
	wcex.hCursor		= LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground	= (HBRUSH)(COLOR_WINDOW+1);
	wcex.lpszMenuName	= (LPCSTR)IDC_WIISCANTRAY;
	wcex.lpszClassName	= szWindowClass;
	wcex.hIconSm		= LoadIcon(wcex.hInstance, (LPCTSTR)IDI_WIISCANICON);

	return RegisterClassEx(&wcex);
}

BOOL InitInstance(HINSTANCE hInstance, int nCmdShow)
{
	FUNSTACK;
	hInst = hInstance; // Store instance handle in our global variable
	HWND hWnd = CreateWindow(szWindowClass, szTitle, WS_OVERLAPPEDWINDOW,CW_USEDEFAULT, 0, CW_USEDEFAULT, 0, NULL, NULL, hInstance, NULL);   
	if (!hWnd) return FALSE;

	// Create the tray icon
	if (!g_trayicon.Create(hInstance,hWnd,WM_ICON_NOTIFY,"Wiiscan monitor",::LoadIcon(hInstance,(LPCTSTR)IDI_WIISCANICON),IDR_POPUP_MENU)) return FALSE;
	g_trayicon.SetIcon(IDI_WIISCANICON_INV);
	
	//ShowWindow(hWnd, nCmdShow);
	UpdateWindow(hWnd);

	return TRUE;
}

int APIENTRY WinMain(HINSTANCE hInstance,HINSTANCE hPrevInstance,LPSTR lpCmdLine,int nCmdShow)
{
	FUNSTACK;
	try{
		int retries=20,n;
		do{
			if (retries!=20) Sleep(100);
			const DWORD id=GetCurrentProcessId();
			const PROCESSENTRY32 p32=GetProcessInfo(id);
			n=GetProcessCount(p32.szExeFile);
		} while(n>1 && --retries>0);

		if (n>1) {
			Message("Error...","Wiiscantray is already running\n\nAborting...");
			return -1;
		}

		// Initialize global strings
		LoadString(hInstance, IDS_APP_TITLE, szTitle, MAX_LOADSTRING);
		LoadString(hInstance, IDC_WIISCANTRAY, szWindowClass, MAX_LOADSTRING);
		MyRegisterClass(hInstance);

		// Perform application initialization:
		if (!InitInstance (hInstance, nCmdShow)) return FALSE;

		HACCEL hAccelTable = LoadAccelerators(hInstance, (LPCTSTR)IDC_WIISCANTRAY);
		
		DWORD tid=0;
		DeviceAutoClose<HANDLE,BOOL> t(CreateThread(0,0,AutoIcon,0,0,&tid),CloseHandle);
		if (t()==NULL) throw_("Error creating autoicon thread");		


		// Main message loop:
		MSG msg;
		while (GetMessage(&msg, NULL, 0, 0)) {
			if (!TranslateAccelerator(msg.hwnd, hAccelTable, &msg)) {
				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}
		}

		return msg.wParam;
	}
	CATCH_ALL;

	return -1;
}